#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netdb.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include "ptp.h"
#include "ptpcommon.h"
#include "ptpraw.h"

#define	PTP_RECV_BUFFER_SIZE	2048
#define	PTP_HANDLER_BACKLOG	10

#define	MAX(a, b)	(a) > (b) ? (a) : (b)

typedef struct	_packet_list {
	int			len;
	__u16			port;
	__u32			addr;
	unsigned char		*packet;
	struct _packet_list	*prev, *next;
} packet_list;

typedef struct _client_list {
	int			fd, buffer_idx;
	unsigned char		buffer[sizeof(__u16) + sizeof(__u32)];
	__u16			req_port;
	__u32			req_addr;
 	struct _client_list	*prev, *next;
} client_list;

unsigned long	g_address = 0;
packet_list	*g_packets = NULL;
client_list	*g_clients = NULL;

int install_ptp(void)
{
	int			fd, on = 1;
	struct sockaddr_in	sin;
	
	if((fd = socket(AF_INET, SOCK_RAW, PTP_PROTO)) == -1) {
		perror("\nsocket");
		return(-1);
	}
	
	if(setsockopt(fd, IPPROTO_IP, IP_HDRINCL, (char *)&on, sizeof(on)) == -1) {
		close(fd);
		perror("\nsetsockopt");
		return(-1);
	}

	memset(&sin, 0x00, sizeof(struct sockaddr_in));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = g_address;

	if(bind(fd, (struct sockaddr *)&sin, sizeof(struct sockaddr)) == -1) {
		close(fd);
		perror("\nbind");
		return(-1);
	}
		
	return(fd);
}

int install_ipc_uds(char *path)
{
	int			fd, len;
	struct sockaddr_un	sun;
	
	if((fd = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
		perror("\nsocket");
		return(-1);
	}
	
	sun.sun_family = AF_UNIX;
	strcpy(sun.sun_path, path);
	len = sizeof(sun);

	if(connect(fd, (struct sockaddr *)&sun, len) != -1) {
        	printf("\nthis service is already running\n");
		exit(1);
	}
       
       	unlink(sun.sun_path);
	
	if(bind(fd, (struct sockaddr *)&sun, len) == -1) {
		perror("\nbind");
		return(-1);
	}

	if(listen(fd, PTP_HANDLER_BACKLOG) == -1) {
		perror("\nlisten");
		return(-1);
	}

	return(fd);
}

int query_device(char *device)
{
	int		if_fd;
	struct ifreq	if_req;
	
	if_fd = socket(AF_INET, SOCK_DGRAM, 0);
	if(if_fd) {
		strcpy(if_req.ifr_name, device);
		if_req.ifr_addr.sa_family = AF_INET;
		if(ioctl(if_fd, SIOCGIFADDR, &if_req) >= 0)
			g_address = ((struct sockaddr_in *)&if_req.ifr_addr)->sin_addr.s_addr;
		else {
			perror("\nioctl");
			close(if_fd);
			return(0);
		}
		
		close(if_fd);
	}
	else {
		perror("\nsocket");
		return(0);
	}
	
	return(1);
}

void exclude_packet(packet_list *packet)
{
	if(packet->prev)
		packet->prev->next = packet->next;
	else
		g_packets = packet->next;

	if(packet->next)
		packet->next->prev = packet->prev;

	free(packet);
}

client_list *locate_client(__u16 port, __u32 addr)
{
	client_list	*temp_cl;
	
	temp_cl = g_clients;
	while(temp_cl) {
		if((port == temp_cl->req_port) && (addr == temp_cl->req_addr))
			return(temp_cl);
		else
			temp_cl = temp_cl->next;			
	}
	
	return(NULL);
}

void cleanup_packets(void)
{
	packet_list	*temp_pl, *temp_pl1;
	
	// removes queued packets which address/port combinations aren't requested by any client anymore
	temp_pl = g_packets;
	while(temp_pl) {
		temp_pl1 = temp_pl->next;

		if(!locate_client(temp_pl->port, temp_pl->addr))
			exclude_packet(temp_pl);
			
		temp_pl = temp_pl1;
	}
}

void exclude_client(client_list *client, fd_set *read_set, fd_set *err_set)
{
	FD_CLR(client->fd, read_set);
	FD_CLR(client->fd, err_set);

	if(client->prev)
		client->prev->next = client->next;
	else
		g_clients = client->next;

	if(client->next)
		client->next->prev = client->prev;

	free(client);
	
	cleanup_packets();
}

packet_list *locate_packet(__u16 port, __u32 addr)
{
	int		found = 0;
	packet_list	*temp_pl;
	
	temp_pl = g_packets;
	while(!found && temp_pl) {
		if((temp_pl->port == port) && (temp_pl->addr == addr))
			found = 1;
		else
			temp_pl = temp_pl->next;
	}

	if(found)
		return(temp_pl);
	else
		return(NULL);
}

void enqueue_packet(int fd_ptp)
{
	int			packet_len, ip_len;
	unsigned char		recv_buffer[PTP_RECV_BUFFER_SIZE];
	socklen_t		sin_len = sizeof(struct sockaddr_in);
	struct in_addr		inaddr;
	struct sockaddr_in	sin;
	struct iphdr		*ip_hdr;
	struct ptphdr		*ptp_hdr;
	packet_list		*temp_pl, *new_pl;

	memset(&sin, 0x00, sizeof(struct sockaddr_in));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = g_address;

	packet_len = recvfrom(fd_ptp, recv_buffer, PTP_RECV_BUFFER_SIZE, 0x00, (struct sockaddr *)&sin, &sin_len);
	if(packet_len < 0) {
		perror("\nrecvfrom (ptp raw)");		// hmm?
		exit(1);
	}
	
	// verify packet
	ip_hdr = (struct iphdr *)recv_buffer;
	ip_len = ip_hdr->ihl * 4;
	
	if((unsigned)packet_len < (ip_len + sizeof(struct ptphdr))) {
_DMSG0("[recv_ptp(): packet too short]\n");
		return;
	}

	ptp_hdr = (struct ptphdr *)(recv_buffer + ip_len);
	ptp_hdr->check = gen_checksum((unsigned short *)ptp_hdr, packet_len - ip_len);
	if(ptp_hdr->check != 0x00) {
_DMSG0("[recv_ptp(): packet checksum mismatch]\n");
		return;
	}

	if(ptp_hdr->type > PTP_TYPE_ADDR_TIME_IFSPEED_IFOCTETS) {
_DMSG1("[recv_ptp(): unknown PTP type %d]\n", ptp_hdr->type);
		return;
	}

	// enqueue packet, including IP header
	inaddr.s_addr = ip_hdr->saddr;
	if(!locate_client(ptp_hdr->dest, ip_hdr->saddr)) {
		printf("(dropped new PTP packet, source %s, port %d)\n", inet_ntoa(inaddr), ptp_hdr->dest);
		return;		// no client known to use this packet, drop it (real PTP would do the same, no caching!)
	}
	
	new_pl = new(packet_list);
	new_pl->prev = new_pl->next = NULL;
	new_pl->port = ptp_hdr->dest;			// for quick selection afterwards
	new_pl->addr = ip_hdr->saddr;			// -''-
	new_pl->len = packet_len;
	new_pl->packet = new(unsigned char[packet_len]);
	memcpy(new_pl->packet, recv_buffer, packet_len);

	temp_pl = g_packets;
	if(temp_pl) {
		while(temp_pl->next)
			temp_pl = temp_pl->next;
		
		temp_pl->next = new_pl;
		new_pl->prev = temp_pl;
	}
	else
		g_packets = new_pl;
		
	printf("(enqueued new PTP packet, source %s, port %d)\n", inet_ntoa(inaddr), ptp_hdr->dest);
}

int handle_query(client_list *client)
{
	int		len;
	__u32		packet_len;
	packet_list	*temp_pl;
	struct in_addr	inaddr;
	
	// to make things simple, a query is just a 16bit number (the PTP-port) followed by a 32bit address (source-address)
	// and we return the oldest packet matching them
	// (32bit byte-count, followed by the binary data)
	temp_pl = locate_packet(client->req_port, client->req_addr);
	if(temp_pl) {
		packet_len = temp_pl->len;
		len = send(client->fd, &packet_len, sizeof(__u32), 0);
		len += send(client->fd, temp_pl->packet, temp_pl->len, 0);
		if(len != ((signed)sizeof(__u32) + temp_pl->len)) {
			// crap, exclude client (we save this packet)
			return(0);
		}
		
		inaddr.s_addr = temp_pl->addr;
		printf("(delivered PTP packet with port %d from address %s)\n", temp_pl->port, inet_ntoa(inaddr));
		exclude_packet(temp_pl);
	}
	
	return(1);
}

void do_handler(int fd_uds, int fd_ptp)
{
	int			sel_ret, fd_max, fd_new, len;
	socklen_t		sock_len;
	fd_set			read_set, read_tmp, err_set, err_tmp;
	client_list		*temp_cl, *temp_cl1, *new_cl;
	struct in_addr		inaddr;
	struct sockaddr_un	sun;
	
	fd_max = MAX(fd_uds, fd_ptp);
	
	FD_ZERO(&read_set);
	FD_ZERO(&err_set);

	FD_SET(fd_uds, &read_set);	// to accept new local ptp queries
	FD_SET(fd_uds, &err_set);

	FD_SET(fd_ptp, &read_set);	// to read raw ptp packets
	FD_SET(fd_ptp, &err_set);
	
	memcpy(&read_tmp, &read_set, sizeof(read_set));	// backup fd selection
	memcpy(&err_tmp, &err_set, sizeof(err_set));
	
	while(1) {
		sel_ret = select(fd_max + 1, &read_set, NULL, &err_set, NULL);
		
		if(sel_ret < 1) {
			perror("\nselect");	// some major error
			exit(1);
		}
		
		// look what fd has something new
		if(FD_ISSET(fd_uds, &read_set)) {
			// new ipc connection to accept
			sock_len = sizeof(sun);
			if((fd_new = accept(fd_uds, (struct sockaddr *)&sun, &sock_len)) == -1) {
				perror("\naccept (IPC UDS)");	// hmm..
				exit(1);
			}
			
			// enqueue this client
			new_cl = new(client_list);
			new_cl->prev = new_cl->next = NULL;
			new_cl->fd = fd_new;
			new_cl->buffer_idx = 0;
			
			temp_cl = g_clients;
			if(temp_cl) {
				while(temp_cl->next)
					temp_cl = temp_cl->next;
				
				temp_cl->next = new_cl;
				new_cl->prev = temp_cl;
			}
			else
				g_clients = new_cl;
			
			// add fd to backup select set
			FD_SET(fd_new, &read_tmp);
			FD_SET(fd_new, &err_tmp);
			
			fd_max = MAX(fd_max, fd_new);

			printf("(accepted new local client)\n");
		}
		
		// check clients for requests
		temp_cl = g_clients;
		while(temp_cl) {
			temp_cl1 = temp_cl->next;
			if(FD_ISSET(temp_cl->fd, &read_set)) {
				// client sent the request (we still read at most the size of the PTP port/address request)
				len = recv(temp_cl->fd, temp_cl->buffer + temp_cl->buffer_idx, sizeof(__u16) + sizeof(__u32) - temp_cl->buffer_idx, 0);
				if(len <= 0) {
					close(temp_cl->fd);
					exclude_client(temp_cl, &read_tmp, &err_tmp);

					printf("(excluded a local client)\n");
				}
				else if((temp_cl->buffer_idx + len) >= ((signed)sizeof(__u16) + (signed)sizeof(__u32))) {
					if(temp_cl->buffer_idx < ((signed)sizeof(__u16) + (signed)sizeof(__u32))) {
						temp_cl->req_port = *((__u16*)temp_cl->buffer);
						temp_cl->req_addr = *((__u32*)(temp_cl->buffer + sizeof(__u16)));
						
						inaddr.s_addr = temp_cl->req_addr;	
						printf("(received PTP port specification, client wants port %d from address %s)\n", temp_cl->req_port, inet_ntoa(inaddr));
					}
						
					temp_cl->buffer_idx += len;
				}
			}
			else if(FD_ISSET(temp_cl->fd, &err_set)) {
				// disconnect this client
				close(temp_cl->fd);
				exclude_client(temp_cl, &read_tmp, &err_tmp);
				
				printf("(excluded a local client)\n");
			}
			
			temp_cl = temp_cl1;
		}

		if(FD_ISSET(fd_ptp, &read_set)) {
			// new raw ptp packet arrived
			enqueue_packet(fd_ptp);
			
			// since we got a new packet, deliver it to the existing clients (if anyone wants...)
			temp_cl = g_clients;
			while(temp_cl) {
				temp_cl1 = temp_cl->next;
				
				if(temp_cl->buffer_idx >= ((signed)sizeof(__u16) + (signed)sizeof(__u32))) {
					if(!handle_query(temp_cl)) {
						close(temp_cl->fd);
						exclude_client(temp_cl, &read_tmp, &err_tmp);

						printf("(excluded a local client)\n");
					}
				}
				temp_cl = temp_cl1;
			}
		}

		if(FD_ISSET(fd_uds, &err_set)) {
			// ipc listener died.. strange
			printf("\nIPC UDS listener died\n");
			exit(1);
		}
		
		if(FD_ISSET(fd_ptp, &err_set)) {
			// raw ptp receiver died.. strange
			printf("\nraw PTP receiver died\n");
			exit(1);
		}
		
		// restore fd selection
		memcpy(&read_set, &read_tmp, sizeof(read_set));
		memcpy(&err_set, &err_tmp, sizeof(err_set));
	}
}

int main(int argc, char **argv)
{
	int		fd_uds, fd_ptp;
	
	if(argc != 2) {
		printf("usage: handler network-device\n");
		exit(1);
	}
	
	printf("determining network device address...");
	fflush(stdout);
	if(!query_device(argv[1]))
		exit(1);
		
	printf("ok\ninstalling IPC unix domain socket...");
	fflush(stdout);
	if((fd_uds = install_ipc_uds(PTP_UNIX_SOCKET_PATH)) == -1)
		exit(1);
	
	printf("ok\ninstalling PTP listener...");
	fflush(stdout);
	if((fd_ptp = install_ptp()) == -1) {
		close(fd_uds);
		exit(1);
	}
	
	printf("ok\n\nrunning (terminate with CTRL-C)\n");
	
	do_handler(fd_uds, fd_ptp);
       	return(0);
}
