#ifndef	__PTPCLIENT_H
#define	__PTPCLIENT_H

#include <linux/types.h>
#include <pthread.h>
#include <netinet/in.h>
#include "ptpcommon.h"
#include "ptpraw.h"

// the various states which can be returned by get_state(NULL)
#define	PTP_STATE_STOPPED	0
#define	PTP_STATE_RUNNING	1

// the various error codes which can be set by get_state(int *error_code)
#define	PTP_ERR_NONE		1000
#define	PTP_ERR_CREATESOCKET	1001
#define	PTP_ERR_NOSUCHDEVICE	1002
#define	PTP_ERR_HOSTLOOKUP	1003
#define	PTP_ERR_CREATETHREAD	1004
#define	PTP_ERR_INITRAW		1005
#define	PTP_ERR_RECVRAW		1006
#define	PTP_ERR_SENDRAW		1007

typedef struct _ptpAddressTable {
	__u32			address, ifspeed, time_old, time_new, ifoctets_old, ifoctets_new;
	struct _ptpAddressTable	*next;
} ptpAddressTable;

class PTPclient {
private:
	int		is_init, state, error_code, ptp_address_entries, ptp_bandwidth_ok;
	__u32		ptp_min_bandwidth;
	unsigned long	device;
	pthread_t	ptp_thread;
	pthread_attr_t	ptp_attr;
	PTPraw		ptp_raw;
	ptpAddressTable	*ptp_address_table;
	
	void process_ifmtu(const unsigned char *packet, int len);
	void process_ifspeed(const unsigned char *packet, int len);
	void process_addr_ifspeed(const unsigned char *packet, int len);
	void process_addr_time_ifoctets(const unsigned char *packet, int len);
	void process_addr_time_ifspeed_ifoctets(const unsigned char *packet, int len);

	void update_address_table(int skip_fields, int ds_count, const __u32 *data);
	void handle_bandwidth_measurement(int ds_count, const __u32 *data, __u8 ptp_type, __u32 src_addr, __u8 ttl);
		
public:
	PTPclient();
	~PTPclient();
	
	// init() is the first function to call. It spawns off the measurement
	// thread, bound to the given device.
	// remote is the IP/DNS of the remote PTPserver (it will just see packets from this address, matching the given port)
	// Returns nonzero on success.
	// (multiple calls will issue a shutdown() between)
	int init(const char *device, const char *remote, __u16 ptp_port);
	
	// shutdown() terminates the thread and shuts down PTP
	void shutdown(void);

	// get_state() returns the current state. If error is not NULL, it will set it accordingly to the most recent error
	int get_state(int *error_code);

	// never call this (must be public for thread workaround)
	void thread_run(void);
};

#endif
