#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/un.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include "ptp.h"
#include "ptpraw.h"
#include "ptpcommon.h"

unsigned short gen_checksum(unsigned short *buff, int len)
{
	register int		sum = 0;
	unsigned short		ret = 0;
	register unsigned short	*w = buff;
	register int		nleft = len;

	while(nleft > 1)  {
		sum += *w++;
		nleft -= 2;
	}

	if(nleft == 1) {
		*(unsigned char *)(&ret) = *(unsigned char *)w ;
		sum += ret;
	}

	sum = (sum >> 16) + (sum & 0xffff);	// add hi 16 to low 16
	sum += (sum >> 16);			// add carry
	ret = ~sum;				// truncate to 16 bits

	return(ret);
}

PTPraw::PTPraw()
{
	this->uds_connected = 0;
	this->sock_fd = this->uds_fd = -1;
	this->ip_id = 0;
}

PTPraw::~PTPraw()
{
	if(this->uds_connected)
		close(this->uds_fd);
	
	if(this->sock_fd != -1)
		close(this->sock_fd);
}

int PTPraw::connect_uds(void)
{
	int			len;
	unsigned char		specs[sizeof(__u16) + sizeof(__u32)];
	struct sockaddr_un	sun;

	if(this->uds_connected) {
		close(this->uds_fd);
		this->uds_connected = 0;
	}

	if((this->uds_fd = socket(AF_UNIX, SOCK_STREAM, 0)) == -1)
		return(0);
		
	sun.sun_family = AF_UNIX;
	strcpy(sun.sun_path, PTP_UNIX_SOCKET_PATH);
	len = sizeof(sun);

	if(connect(this->uds_fd, (struct sockaddr *)&sun, len) == -1) {
		close(this->uds_fd);
		return(0);
	}

	// immediately send PTP port/address identifier
	memcpy(specs, &this->ptp_port, sizeof(__u16));
	memcpy(specs + sizeof(__u16), &this->remote_addr.s_addr, sizeof(__u32));

	if(send(this->uds_fd, specs, sizeof(__u16) + sizeof(__u32), 0) != (sizeof(__u16) + sizeof(__u32))) {
		close(this->uds_fd);
		return(0);
	}
	
	this->uds_connected = 1;
	return(1);
}

int PTPraw::init(struct in_addr local_addr, struct in_addr remote_addr, __u16 ptp_port)
{
	int			on = 1;
	struct sockaddr_in	sin;

	this->local_addr = local_addr;
	this->remote_addr = remote_addr;
	this->ptp_port = ptp_port;

	// setup IPC unix domain socket to recv packets from PTP handler
	if(!this->connect_uds())
		return(0);

	// setup raw socket to send packets
	if((this->sock_fd = socket(AF_INET, SOCK_RAW, PTP_PROTO)) == -1)
		return(0);

	if(setsockopt(this->sock_fd, IPPROTO_IP, IP_HDRINCL, (char *)&on, sizeof(on)) == -1) {
		close(this->sock_fd);
		return(0);
	}

	memset(&sin, 0x00, sizeof(struct sockaddr_in));
	sin.sin_family = AF_INET;
	sin.sin_addr = this->local_addr;

	if(bind(this->sock_fd, (struct sockaddr *)&sin, sizeof(struct sockaddr)) == -1) {
		close(this->sock_fd);
		return(0);
	}

	return(1);
}

int PTPraw::recv_ptp(long timeout, unsigned char **ptp_packet)
{
	int		result;
	__u32		packet_len;
	fd_set		read_set, err_set;
	struct timeval	tv;

	if(!this->uds_connected) {
		if(!this->connect_uds())
			return(0);
	}

	FD_ZERO(&read_set);
	FD_SET(this->uds_fd, &read_set);
	FD_ZERO(&err_set);
	FD_SET(this->uds_fd, &err_set);

	if(timeout != -1) {
		tv.tv_sec = timeout / 1000;
		tv.tv_usec = (timeout - tv.tv_sec * 1000) * 1000;
		result = select(this->uds_fd + 1, &read_set, NULL, &err_set, &tv);
	}
	else
		result = select(this->uds_fd + 1, &read_set, NULL, &err_set, NULL);

	if(result < 1) {
		// nothing to read yet or error
		if(result < 0)
			this->connect_uds();

		return(0);
	}

	if(FD_ISSET(this->uds_fd, &err_set)) {
		// this is pretty bad now, try to reconnect to uds
		this->connect_uds();
		return(0);
	}

	if(recv(this->uds_fd, &packet_len, sizeof(__u32), 0) != sizeof(__u32)) {
		// we couldn't even read the packet length, try to reconnect to uds
		this->connect_uds();
		return(0);
	}

	*ptp_packet = new(unsigned char[packet_len]);
	if(recv(this->uds_fd, *ptp_packet, packet_len, 0) != (signed)packet_len) {
		// we couldn't read the packet, try to reconnect to uds
		delete(*ptp_packet);
		this->connect_uds();
		return(0);
	}

	return(packet_len);
}

int PTPraw::send_ptp(int packet_length, const unsigned char *packet)
{
	struct iphdr		*ip_hdr;
	struct sockaddr_in	sin;

	ip_hdr = (struct iphdr *)packet;

	memset(&sin, 0x00, sizeof(struct sockaddr_in));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = ip_hdr->daddr;

	if(sendto(this->sock_fd, packet, packet_length, 0x0, (struct sockaddr *)&sin, sizeof(struct sockaddr_in)) != packet_length)
		return(0);

	return(1);
}

unsigned char *PTPraw::assemble_ptp(__u8 ttl, __u32 daddr, __u8 ttl_check, __u8 compare_f, __u8 echo_f, __u8 ds_len, __u8 ds_count, int content_type, int data_len, int check_len, const unsigned char *data, int *packet_len)
{
	int			ptp_offset;
	__u32			router_alert = 0;//IP_OPTION_ROUTER_ALERT;
	unsigned char		*packet;
	struct iphdr		*ip_hdr;
	struct ptphdr		*ptp_hdr;

	if(data_len < check_len)
		return(NULL);

	switch(content_type) {
	case PTP_TYPE_IFMTU:
	case PTP_TYPE_IFSPEED:
		if(ds_len % 2)
			return(NULL);
		break;

	case PTP_TYPE_ADDR_IFSPEED:
		//if(ds_len % 4) - doesn't work for response packets :/
		//	return(NULL);

		if(compare_f)
			return(NULL);
		break;

	case PTP_TYPE_ADDR_TIME_IFOCTETS:
		//if(ds_len % 6) - doesn't work for response packets :/
		//	return(NULL);

		if(compare_f)
			return(NULL);
		break;

	case PTP_TYPE_ADDR_TIME_IFSPEED_IFOCTETS:
		//if(ds_len % 8) - doesn't work for response packets :/
		//	return(NULL);

		if(compare_f)
			return(NULL);
		break;

	default:
		return(NULL);
	}

	*packet_len = sizeof(struct iphdr) + 4 + sizeof(struct ptphdr) + data_len;
	ptp_offset = sizeof(struct iphdr) + 4;

	packet = new(unsigned char[*packet_len]);
	if(!packet)
		return(NULL);

	memset(packet, 0, *packet_len);

	// fill in IP header
	ip_hdr = (struct iphdr *)packet;

	ip_hdr->ihl = 6;			// 5 dwords iphdr len + 1 IP option
	ip_hdr->version = 4;			// good ol' IPv4
	ip_hdr->tos = 0;			// ignored nowadays
	ip_hdr->tot_len = htons(*packet_len);

	ip_hdr->id = htons(this->ip_id++);	// you never know...
	ip_hdr->frag_off = 0x40;		// don't fragment
	ip_hdr->ttl = ttl;
	ip_hdr->protocol = PTP_PROTO;

	ip_hdr->saddr = this->local_addr.s_addr;
	ip_hdr->daddr = daddr;

	memcpy(packet + sizeof(struct iphdr), &router_alert, 4);

	ip_hdr->check = 0;
	ip_hdr->check = gen_checksum((unsigned short *)ip_hdr, sizeof(struct iphdr) + 4);

	// fill in PTP header
	ptp_hdr = (struct ptphdr *)(packet + ptp_offset);

	ptp_hdr->ttl_check = ttl_check;
	ptp_hdr->length = ds_len;
	ptp_hdr->compare = compare_f;
	ptp_hdr->echo = echo_f;
	ptp_hdr->type = content_type;
	ptp_hdr->dest = this->ptp_port;
	ptp_hdr->count = ds_count;

	memcpy(packet + ptp_offset + sizeof(struct ptphdr), data, data_len);

	ptp_hdr->check = 0;
	ptp_hdr->check = gen_checksum((unsigned short *)ptp_hdr, sizeof(struct ptphdr) + check_len);

	return(packet);
}

