#ifndef	__PTP_RAW_H
#define	__PTP_RAW_H

#include <linux/types.h>

#define	PTP_UNIX_SOCKET_PATH	"/tmp/.ptphandlersock"

unsigned short gen_checksum(unsigned short *buff, int len);

class PTPraw {
private:
	int		sock_fd, uds_fd, uds_connected;
	__u16		ip_id, ptp_port;
	struct in_addr	local_addr, remote_addr;

	int connect_uds(void);

public:
	PTPraw();
	~PTPraw();

	// the first function to call, which specifies to what interface the packets are bound,
	// what remote address they need to have to get returned by recv_ptp, and what port to use
	int init(struct in_addr local_addr, struct in_addr remote_addr, __u16 ptp_port);

	// receives a single PTP packet and returns it with all headers (IP and PTP)
	// timeout is in milliseconds or -1 to block (this CAN use more time than the given one!)
	// this will just return ptp packets which pass certain general checks, matching the given port
	// returns packet length on success, 0 on error
	int recv_ptp(long timeout, unsigned char **ptp_packet);

	// sends a single fully-assembled (IP + PTP header + PTP data) PTP packet (use assemble_ptp to assemble this one)
	// returns nonzero on success
	int send_ptp(int packet_len, const unsigned char *packet);

	// assembles a full-featured PTP packet, including the wrapping IP header and all checksums calculated
	// packet_len is set to the final length of the returned packet
	// performs some sanity checks and returns NULL upon error
	unsigned char *assemble_ptp(__u8 ttl, __u32 daddr, __u8 ttl_check, __u8 compare_f, __u8 echo_f, __u8 ds_len, __u8 ds_count, int content_type, int data_len, int check_len, const unsigned char *data, int *packet_len);
	unsigned char *assemble_ptp2(__u8 ttl, __u32 daddr, __u8 ttl_check, __u8 compare_f, __u8 echo_f, __u8 ds_len, __u8 ds_count, int content_type, int data_len, int check_len, struct ptpmsg *ptpdata, int *packet_len);

};

#endif
