#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include "ptpserver.h"

void print_usage(void)
{
	printf("\n(1) get_path_mtu(c=0, e=0, 0 PDU space)\n");
	printf("(2) get_path_mtu(c=1, e=0, min_mtu=0xffff)\n");
	printf("(3) get_ifspeed(c=0, e=0, 0 PDU space)\n");
	printf("(4) get_ifspeed(c=1, e=0, min_ifspeed=0xffffffff\n");
	printf("(5) custom selection of the above\n");
	printf("(6) launch bandwidth measurement thread (FOR NOW: only with type 4 packets, 0 PDU space)\n");
	printf("(7) get bandwidth status\n");
	printf("(8) shutdown bandwidth measurement thread\n\n? ");
	fflush(stdout);
}

int select_custom(int *comp, int *echo, __u32 *min, int *pdu)
{
	char	c;

	printf("- (a) get_path_mtu()\n");
	printf("- (b) get_ifspeed()\n");
	printf("- (x) return\n\n? ");
	fflush(stdout);

	do {
		scanf("%c", &c);
		if(islower(c))
			c = toupper(c);

		if(c == 'A') {
			printf("- get_path_mtu: c e min pdu ? ");
			scanf("%d %d %u %d", comp, echo, min, pdu);
			return(1);
		}
		else if(c == 'B') {
			printf("- get_ifspeed: c e min pdu ? ");
			scanf("%d %d %u %d", comp, echo, min, pdu);
			return(2);
		}
		else if(c == 'X')
			return(0);
	} while(1);
}

int main(int argc, char **argv)
{
	int		pu, compare, echo, pdu, type, err, route, counter;
	char		ch;
	__u8		missing;
	__u32		mtu, ifspeed, min, interval, mib2_ifspeed, traffic;
	PTPserver	server;

	if(argc != 5) {
		printf("usage: %s device destination rtt port\n", argv[0]);
		return(1);
	}

	if(!server.init(argv[1], argv[2], atoi(argv[3]), atoi(argv[4]))) {
		printf("init: %d\n", server.get_error());
		return(1);
	}

	printf("[CTRL-C to abort]\n\n");

	print_usage();

	do {
		pu = 1;

		scanf("%c", &ch);
		switch(atoi(&ch)) {
		case 1:
			printf("* get_path_mtu(c=0, e=0, 0 PDU space):\n");
			mtu = server.get_path_mtu(0, 0, 0, 0, &missing);
			if(mtu == 0)
				printf("* error: %d\n", server.get_error());
			else
				printf("* success, mtu: %d (missing links: %d)\n", mtu, missing);
			break;

		case 2:
			printf("* get_path_mtu(c=1, e=0, min_mtu=0xffff):\n");
			mtu = server.get_path_mtu(1, 0, 0xffff, 0, &missing);
			if(mtu == 0)
				printf("*error: %d\n", server.get_error());
			else
				printf("*success, mtu: %d (missing links: %d)\n", mtu, missing);
			break;

		case 3:
			printf("*get_ifspeed(c=0, e=0, 0 PDU space):\n");
			ifspeed = server.get_ifspeed(0, 0, 0, 0, &missing);
			if(ifspeed == 0)
				printf("*error: %d\n", server.get_error());
			else
				printf("*success, ifspeed: %d (missing links: %d)\n", ifspeed, missing);
			break;

		case 4:
			printf("* get_ifspeed(c=1, e=0, min_ifspeed=0xffffffff):\n");
			ifspeed = server.get_ifspeed(1, 0, 0xffffffff, 0, &missing);
			if(ifspeed == 0)
				printf("* error: %d\n", server.get_error());
			else
				printf("* success, ifspeed: %d (missing links: %d)\n", ifspeed, missing);
			break;

		case 5:
			if((type = select_custom(&compare, &echo, &min, &pdu))) {
				if(type == 1) {
					if(compare)
						printf("* get_path_mtu(c=%d, e=%d, min_mtu=%u):\n", compare, echo, min);
					else
						printf("* get_path_mtu(c=%d, e=%d, %d PDU space):\n", compare, echo, pdu);

					mtu = server.get_path_mtu(compare, echo, compare ? min : 0, compare ? 0 : pdu, &missing);
					if(mtu == 0)
						printf("* error: %d\n", server.get_error());
					else
						printf("* success, mtu: %d (missing links: %d)\n", mtu, missing);
				}
				else {
					if(compare)
						printf("* get_ifspeed(c=%d, e=%d, min_ifspeed=%u):\n", compare, echo, min);
					else
						printf("* get_ifspeed(c=%d, e=%d, %d PDU space):\n", compare, echo, pdu);

					ifspeed = server.get_ifspeed(compare, echo, compare ? min : 0, compare ? 0 : pdu, &missing);
					if(ifspeed == 0)
						printf("* error: %d\n", server.get_error());
					else
						printf("* success, ifspeed: %d (missing links: %d)\n", ifspeed, missing);
				}
			}
			break;

		case 6:
			if(!server.start_availbw_detection(PTP_BANDWIDTH_TYPE_4, 0))
				printf("* error: %d\n", server.get_error());
			else
				printf("* success, bandwidth measurement thread is running\n");
			break;

		case 7:
			err = server.get_bandwidth(&route, &counter, &interval, &mib2_ifspeed, &traffic);
			if(err != PTP_BANDWIDTH_ERR_NONE)
				printf("* error: %d\n", err);
			else {
				printf("* SUCCESS: \n route = %d\n counter = %d\n interval = %d\n mib2_ifspeed = %d\n traffic = %d\n", route, counter, interval, mib2_ifspeed, traffic);
			}
			break;

		case 8:
			server.stop_availbw_detection();
			printf("* thread stopped\n");
			break;

		default:
			pu = 0;
			break;
		}

		if(pu)
			print_usage();
	} while(1);
}
