#include <stdio.h>
#include <ctype.h>
#include "ptpserver.h"
#include "tcp2ptp.h"

void err_dump(char *s) {
  fprintf(stderr,": %s\n",s);
  exit(1);
}

int main(int argc, char **argv) {
  int err, listenfd, connfd, route, counter;
  __u32	interval, mib2_ifspeed, traffic;
  PTPserver server;
  pid_t childpid;
  socklen_t clilen;
  struct sockaddr_in cliaddr, servaddr;
  void sig_chld(int);
  char ch;

  if(argc != 5) {
    printf("usage: %s device destination rtt port\n", argv[0]);
    return(1);
  }

  if(!server.init(argv[1], argv[2], atoi(argv[3]), atoi(argv[4]))) {
    printf("init: %d\n", server.get_error());
    return(1);
  }

  listenfd = socket(AF_INET, SOCK_STREAM, 0);

  bzero(&servaddr, sizeof(servaddr));
  servaddr.sin_family      = AF_INET;
  servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  servaddr.sin_port        = htons(SERV_PORT);

  bind(listenfd, (SA *) &servaddr, sizeof(servaddr));

  listen(listenfd, LISTENQ);

  for ( ; ; ) {
    clilen = sizeof(cliaddr);
    if ( (connfd = accept(listenfd, (SA *) &cliaddr, &clilen)) < 0) {
      if (errno == EINTR)
	continue;
      else
	err_dump("accept error");
    }

    if ( (childpid = fork()) == 0) {
      close(listenfd);

      while( recv( connfd, &ch, 1, 0)) {
	if( atoi( &ch) == 1) {
	  if(!server.start_availbw_detection(PTP_BANDWIDTH_TYPE_4, 0)) {
	    send( connfd, "1", 1, 0);
	    printf("* error: %d\n", server.get_error());
	  }
	  else {
	    send( connfd, "0", 1, 0);
	    printf("* success, bandwidth measurement thread is running\n");
	  }
	}
	else if( atoi( &ch) == 2) {
	  err = server.get_bandwidth(&route, &counter, &interval, &mib2_ifspeed, &traffic);
	  if(err != PTP_BANDWIDTH_ERR_NONE) {
	    send( connfd, "1", 1, 0);
	    printf("* error: %d\n", err);
	  }
//	  else {
	  printf("route = %d, counter = %d, interval = %d, bandwidth = %d, traffic = %d\n", route, counter, interval, mib2_ifspeed, traffic);
//	}
	}
	else if( atoi( &ch) == 3){
	  server.stop_availbw_detection();
	  send( connfd, "0", 1, 0);
	  printf("* bandwidth measurement thread stopped\n");
	  exit(0);
	}
	else
	  exit(0);
      }
    }
    close(connfd);
  }
}












